(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2007,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

open Parsetree;;
open Longident;;

let underscore_pat = { ppat_desc = Ppat_any; ppat_loc = Location.none };;

let rec lower_long_ident = function
  | Lident s -> Lident (String.lowercase s)
  | Ldot (t, s) -> Ldot (t, String.lowercase s)
  | Lapply (t1, t2) -> Lapply (t1, lower_long_ident t2);;

let chg_desc r p = { r with ppat_desc = p };;

let rec lower_case r =
  match r.ppat_desc with
  | Ppat_alias (p, s) -> chg_desc r (Ppat_alias (lower_case p, s))
  | Ppat_tuple pats ->
    chg_desc r (Ppat_tuple (List.map lower_case pats))
  | Ppat_construct (id, Some p, b) ->
    chg_desc r (Ppat_construct (lower_long_ident id, Some (lower_case p), b))
  | Ppat_variant (label, Some p) ->
    chg_desc r (Ppat_variant (label, Some (lower_case p)))
  | Ppat_record idpats ->
    let f (id,pat) = (id, lower_case pat) in
    chg_desc r (Ppat_record (List.map f idpats))
  | Ppat_array pats ->
    chg_desc r (Ppat_array (List.map lower_case pats))
  | Ppat_or (p1, p2) ->
    chg_desc r (Ppat_or (lower_case p1, lower_case p2))
  | _ -> r;;

let remove_topconstr r =
  match r.ppat_desc with
  | Ppat_construct (_, Some p, _) -> p
  | _ -> assert false;;

let rec add_underscores r =
  match r.ppat_desc with
  | Ppat_alias (p, s) -> chg_desc r (Ppat_alias (add_underscores p, s))
  | Ppat_tuple pats ->
    chg_desc r (Ppat_tuple (List.map add_underscores pats))
  | Ppat_construct (Lident s as id, Some p, b) ->
    (try let _ = Check.find_cname s in
       let newp = chg_desc p
         (match p.ppat_desc with
          | Ppat_tuple pats ->
            Ppat_tuple (underscore_pat :: List.map add_underscores pats)
          | _ -> Ppat_tuple [underscore_pat; add_underscores p])
       in chg_desc r (Ppat_construct (id, Some newp, b))
     with Not_found ->
       chg_desc r (Ppat_construct (id, Some (add_underscores p), b)))
  | Ppat_variant (label, Some p) ->
    chg_desc r (Ppat_variant (label, Some (add_underscores p)))
  | Ppat_record idpats ->
    let f (id,pat) = (id, add_underscores pat) in
    chg_desc r (Ppat_record (List.map f idpats))
  | Ppat_array pats ->
    chg_desc r (Ppat_array (List.map add_underscores pats))
  | Ppat_or (p1, p2) ->
    chg_desc r (Ppat_or (add_underscores p1, add_underscores p2))
  | _ -> r;;
