(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2007,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: genr_sharing.ml,v 1.22 2007/04/25 15:04:50 weis Exp $ *)

open Asttypes;;
open Parsetree;;
open Longident;;
open Relation;;
open Spec;;
open Code;;
open Genr_pattern;;
open Otype;;
open Pr_otype;;
open Check;;

let list_map_i f l =
  let rec loop i accu = function
  | [] -> accu
  | x :: l -> let h = f i x in loop (succ i) (h :: accu) l in
  List.rev (loop 0 [] l);;

let initial_sharing_table_size = 1009;;

(* Various names for functions, built from the name of the type. *)

(* The name of the construction function that will share terms. *)
let mk_constr cname = "mk_" ^ cname;;

(* The modules we generate. *)
let hash_module_name tname = "Hash_" ^ tname;;
let share_module_name tname = "Share_" ^ tname;;

(* The name of the type as defined in its private module. *)
let type_name tname = "Type_t." ^ tname;;

(* The functions we generate. *)
let eq_fun_name tname = "eq_" ^ tname;;
let equal_fun_name tname = "equal_" ^ tname;;
let hash_fun_name tname = "hash_" ^ tname;;
let get_hash_fun_name tname = "get_hash_" ^ tname;;

(* The name of the sharing hash table. *)
let hash_table_name tname = "htbl_" ^ tname;;

let get_hash_fun_name_i tname cname i =
  try let ci = Check.find_cname cname in
(* If debugging
   match ci with
   | { ci_type_name = tname;
       ci_arg_types = cts;
       ci_relations = rels} ->
     prerr_endline (Printf.sprintf
       "get_hash_fun_name_i: type %s, constructor %s, arg %i"
       tname cname i);
     prerr_endline (Printf.sprintf
       "length is: %i" (List.length cts));
*)
    let ct = List.nth ci.ci_arg_types (i - 1) in
    (match ct.ptyp_desc with
     | Ptyp_constr (Lident s, _) when s = tname -> get_hash_fun_name tname
     | _ -> "Hashtbl.hash") with
  | Not_found -> "Hashtbl.hash";;

let typ_info = const_type "info";;

let add_info_arg ((cid, tycl, rels, loc) as cstr) =
  match tycl with
  | [] -> cstr
  | _ -> (cid, typ_info :: tycl, rels, loc);;

let remove_cstr_relations (cid, tycl, rels, loc) =
  add_info_arg (cid, tycl, [], loc);;

let make_public_type_kind = function
  | Ptype_abstract (relations, pflag) -> Ptype_abstract (relations, Public)
  | Ptype_variant (cstrs, pflag) ->
      Ptype_variant (List.map remove_cstr_relations cstrs, Public)
  | Ptype_record (flds, pflag) -> Ptype_record (flds, Public);;

let print_typedef b tname =
  let type_decl = find_tname tname in
  let ct = (* Type_t.``tname'' *)
    mktyp
     (Ptyp_constr
       (Ldot (Lident "Type_t", tname),
        List.map var_type type_decl.ptype_params)) in
  let new_type_decl = {
    type_decl with
      ptype_kind = make_public_type_kind type_decl.ptype_kind;
      ptype_manifest = Some ct;
  } in
  Code.structure_item b (Pstr_type [tname, new_type_decl]);;

let find_typedef_params tname =
  let type_decl = find_tname tname in
  type_decl.ptype_params;;

let is_typedef_polymorphic tname = find_typedef_params tname != [];;

let print_type_params b = function
  | [] -> ()
  | [ x ] -> Printf.bprintf b "'%s " x 
  | x :: l ->
    let print_param b x = Printf.bprintf b ", '%s" x in
    let print_params b l = List.iter (print_param b) l in
    Printf.bprintf b "('%s%a) " x print_params l;;

let print_type_shared b tname =
  let ty_vars = find_typedef_params tname in
  Printf.bprintf b "%a%s" print_type_params ty_vars (type_name tname);;

let print_mod_hashedtype b tname =
  let ty_vars = find_typedef_params tname in
  if ty_vars = [] then begin
    (* The monomorphic case. *)
    Printf.bprintf b "\
      \nmodule %s = struct\
      \n  type t = %s;;\
      \n  let equal = %s;;\
      \n  let hash = %s;;\
      \nend;;\n"
      (hash_module_name tname)
      (type_name tname)
      (equal_fun_name tname)
      (hash_fun_name tname);
    Printf.bprintf b "\
      \nmodule %s = Weak.Make (%s);;\n"
      (share_module_name tname) (hash_module_name tname);
    Printf.bprintf b "\
      \nlet %s = %s.create %i;;\n"
      (hash_table_name tname)
      (share_module_name tname)
      initial_sharing_table_size;
  end else begin
    (* The polymorphic case. *)
    let print_equal_fun_body b fname =
      Printf.bprintf b
        "fun x y -> %s (Obj.magic x) (Obj.magic y)"
        fname in
    let print_hash_fun_body b fname =
      Printf.bprintf b
        "fun x -> %s (Obj.magic x : %a)"
        fname
        print_type_shared tname in
    Printf.bprintf b "\
      \nmodule %s = struct\
      \n  type t = %s;;\
      \n  let equal = %a;;\
      \n  let hash = %a;;\
      \nend;;\n"
      (hash_module_name tname)
      "Obj.t"
      print_equal_fun_body (equal_fun_name tname)
      print_hash_fun_body (hash_fun_name tname);
    Printf.bprintf b "\
      \nmodule %s = struct\n\
      \n  module Share = Weak.Make (%s);;\n\
      \n  include Share;;\n\
      \n  let find t x = (Obj.magic (Share.find t (Obj.repr x)));;\
      \n  let add t x = Share.find t (Obj.repr x);;\
      \nend;;\n"
      (share_module_name tname) (hash_module_name tname);
    Printf.bprintf b "\
      \nlet %s = %s.create %i;;\n"
      (hash_table_name tname)
      (share_module_name tname)
      initial_sharing_table_size;
  end;
  print_typedef b tname;;

let constr_underscore (cname, xs) = Constr (cname, underscore :: xs);;

let genr_pattern_underscore cname root n =
  constr_underscore (cname, genr_idargs root n);;

let genr_const_pattern cname = Id cname;;

let constr_info (cname, info, xs) = Constr (cname, info :: xs);;

let genr_pattern_info cname info root n =
  constr_info (cname, info, genr_idargs root n);;

let genr_clause_nullary i cname =
  Clause (genr_const_pattern cname, Id (string_of_int i));;

let genr_clauses clause_nullary clause_non_nullary cds default_clauses =
  (* i is the rank of the current constant (or nullary) constructor.
     j is the rank of the current functional (or non nullary) constructor. *)
  let rec loop (i, j, accu) = function
    | [] -> accu
    | (cname, cts, _rels, _loc) :: cds ->
      match List.length cts with
      | 0 ->
        let accu = clause_nullary i cname accu in
        loop (i + 1, j, accu) cds
      | n ->
        let accu = clause_non_nullary j cname n accu in
        loop (i, j + 1, accu) cds in
  loop (1, 1, default_clauses) cds;;

(* Generates the hash table internal equality function for (sum) type tname
   (the function named equal_tname in the generated source). *)
let genr_equal_sum tname cds =

  let exps_args = [x; y] in
  let args = Code.idents_of_exps exps_args in

  let clause_nullary i cname accu = accu

  and clause_non_nullary j cname n accu =
    let xs = genr_idargs "x" n
    and ys = genr_idargs "y" n in
    let p1 = constr_underscore (cname, xs)
    and p2 = constr_underscore (cname, ys)
    and eqs = List.map2 (fun x y -> appi x "==" y) xs ys in
    let body =
      match eqs with
      | eq :: eqs -> List.fold_left (fun eq e -> appi eq "&&" e) eq eqs
      | _ -> assert false in
    Clause (pair p1 p2, body) :: accu

  and default_clauses = [Clause (underscore, Id "false")] in

  let clauses =
    genr_clauses clause_nullary clause_non_nullary cds default_clauses in
  topletrec [equal_fun_name tname, args, Match (pair x y, clauses)];;

let info_pattern h = Id ("{ hash = " ^ h ^ " }");;

(* Finds the hash code in the tname values, retieved in the info fiels. *)
let genr_get_hash_sum tname cds =

  let exps_args = [x] in
  let args = Code.idents_of_exps exps_args in

  let clause_nullary i cname accu = genr_clause_nullary i cname :: accu

  and clause_non_nullary j cname n accu =
    Clause
     (genr_pattern_info cname (info_pattern "h") "_x" n, Id "h") :: accu

  and default_clauses = [] in

  let clauses =
    genr_clauses clause_nullary clause_non_nullary cds default_clauses in

  topletrec [get_hash_fun_name tname, args, Match (x, clauses)];;

(* Generates the hashing function for (sum) type tname
   (the function named hash_tname in the generated code). *)
let genr_hash_sum tname cds =

  let exps_args = [x] in
  let args = Code.idents_of_exps exps_args in

  let clause_nullary i cname accu = genr_clause_nullary i cname :: accu

  and clause_non_nullary j cname n accu =
    (* The arguments of constructor cname. *)
    let i_xis = genr_numbered_idargs "x" n in
    (* The integer value associated to the constructor, its unique rank in the
    type definition, that is added to the sum of hash codes of the arguments. *)
    let chash = Id (string_of_int j) in
    let sum e (i, xi) =
      appi (app1 (get_hash_fun_name_i tname cname i) xi) "+" e in
    let cl_exp = List.fold_left sum chash i_xis in
    Clause (genr_pattern_underscore cname "x" n, cl_exp) :: accu

  and default_clauses = [] in

  let clauses =
    genr_clauses clause_nullary clause_non_nullary cds default_clauses in

  topletrec [hash_fun_name tname, args, Match (x, clauses)];;

let genr_construction_function tname cname = function
  | 0 -> assert false
  | n ->
    let args = genr_args "x" n and info = Id "info" and v = Id "v"
    and idargs = genr_idargs "x" n and hinfo = Id "info.hash" in
    let modul = share_module_name tname and tbl = Id (hash_table_name tname) in
    let mfind = modul ^ ".find" and madd = modul ^ ".add" in
    let with_clause =
      Clause (Id "Not_found", Let ("_", [], app2 madd tbl v, v)) in

    let body =
      Let ("info", [], Record ["hash", Id "0"],
      Let ("v", [], constr_info (cname, info, idargs),
      Let ("_", [], appi hinfo "<-" (app1 (hash_fun_name tname) v),
      Try (app2 mfind tbl v, [with_clause])))) in

    topletrec [mk_constr cname, args, body];;

let genr_construction_functions tname cds =
  List.map
    (fun (cname, cts, _rels, _loc) ->
      genr_construction_function tname cname (List.length cts))
    (List.filter (fun cd -> cd_arity cd > 0) cds);;

let print_mk_info b =
  Printf.bprintf b "let mk_info h = { hash = h };;\n";;

let print_sharing_preamble = print_mk_info;;

(****
let print_eq_fun_name_definition b tname =
  Printf.bprintf b
   "external %s : %a -> %a -> bool = %S;;\n"
   (eq_fun_name tname) print_type_shared tname print_type_shared tname "%eq";;
*)

let print_sharing b ntds =
  (* FIXME: move from tdef to ocaml tdecl *)
  let print_ntd (tname, td) =
    match td.ptype_kind with
    | Ptype_abstract (_rels, _priv) ->
      failwith "Not yet implemented"
    | Ptype_variant (cds, _priv) ->
      let teq = genr_equal_sum tname cds in
      let tgh = genr_get_hash_sum tname cds in
      let teh = genr_hash_sum tname cds in
      Code.structure b [teq; tgh; teh];
      print_mod_hashedtype b tname;
      Code.structure b (genr_construction_functions tname cds)
    | Ptype_record (fds, _priv) -> failwith "Not yet implemented" in
  List.iter print_ntd ntds;;
